@@ -1,55 +1,59 @@
 /*
  *  socket_linux.c
  *
- *  Copyright 2013-2021 Michael Zillgith
+ *  Copyright 2013-2024 Michael Zillgith
  *
  *  This file is part of Platform Abstraction Layer (libpal)
  *  for libiec61850, libmms, and lib60870.
  */
 
 #include "hal_socket.h"
-#include <sys/types.h>
-#include <sys/socket.h>
-#include <sys/select.h>
 #include <arpa/inet.h>
-#include <unistd.h>
-#include <string.h>
-#include <netinet/in.h>
-#include <netdb.h>
 #include <errno.h>
-#include <stdio.h>
 #include <fcntl.h>
-#include <netinet/tcp.h> /* required for TCP keepalive */
 #include <linux/version.h>
+#include <netdb.h>
+#include <netinet/in.h>
+#include <netinet/tcp.h> /* required for TCP keepalive */
+#include <stdio.h>
+#include <string.h>
+#include <sys/select.h>
+#include <sys/socket.h>
+#include <sys/types.h>
+#include <unistd.h>
 
 #define _GNU_SOURCE
-#include <signal.h>
 #include <poll.h>
+#include <signal.h>
 
-
-#include "linked_list.h"
 #include "hal_thread.h"
 #include "lib_memory.h"
+#include "linked_list.h"
 
 #ifndef DEBUG_SOCKET
 #define DEBUG_SOCKET 0
 #endif
 
-struct sSocket {
+struct sSocket
+{
     int fd;
     uint32_t connectTimeout;
 };
 
-struct sServerSocket {
+struct sServerSocket
+{
     int fd;
     int backLog;
 };
 
-struct sUdpSocket {
+struct sUdpSocket
+{
     int fd;
+    int namespace; /* IPv4: AF_INET; IPv6: AF_INET6 */
 };
 
-struct sHandleSet {
+struct sHandleSet
+{
     LinkedList sockets;
     bool pollfdIsUpdated;
     struct pollfd* fds;
@@ -59,23 +63,26 @@ struct sHandleSet {
 HandleSet
 Handleset_new(void)
 {
-   HandleSet self = (HandleSet) GLOBAL_MALLOC(sizeof(struct sHandleSet));
+    HandleSet self = (HandleSet)GLOBAL_MALLOC(sizeof(struct sHandleSet));
 
-   if (self) {
-       self->sockets = LinkedList_create();
-       self->pollfdIsUpdated = false;
-       self->fds = NULL;
-       self->nfds = 0;
-   }
+    if (self)
+    {
+        self->sockets = LinkedList_create();
+        self->pollfdIsUpdated = false;
+        self->fds = NULL;
+        self->nfds = 0;
+    }
 
-   return self;
+    return self;
 }
 
 void
 Handleset_reset(HandleSet self)
 {
-    if (self) {
-        if (self->sockets) {
+    if (self)
+    {
+        if (self->sockets)
+        {
             LinkedList_destroyStatic(self->sockets);
             self->sockets = LinkedList_create();
             self->pollfdIsUpdated = false;
@@ -86,17 +93,19 @@ Handleset_reset(HandleSet self)
 void
 Handleset_addSocket(HandleSet self, const Socket sock)
 {
-   if (self != NULL && sock != NULL && sock->fd != -1) {
+    if (self != NULL && sock != NULL && sock->fd != -1)
+    {
 
-       LinkedList_add(self->sockets, sock);
-       self->pollfdIsUpdated = false;
-   }
+        LinkedList_add(self->sockets, sock);
+        self->pollfdIsUpdated = false;
+    }
 }
 
 void
 Handleset_removeSocket(HandleSet self, const Socket sock)
 {
-    if (self && self->sockets && sock) {
+    if (self && self->sockets && sock)
+    {
         LinkedList_remove(self->sockets, sock);
         self->pollfdIsUpdated = false;
     }
@@ -106,8 +115,10 @@ int
 Handleset_waitReady(HandleSet self, unsigned int timeoutMs)
 {
     /* check if pollfd array is updated */
-    if (self->pollfdIsUpdated == false) {
-        if (self->fds) {
+    if (self->pollfdIsUpdated == false)
+    {
+        if (self->fds)
+        {
             GLOBAL_FREEMEM(self->fds);
             self->fds = NULL;
         }
@@ -118,13 +129,16 @@ Handleset_waitReady(HandleSet self, unsigned int timeoutMs)
 
         int i;
 
-        for (i = 0; i < self->nfds; i++) {
+        for (i = 0; i < self->nfds; i++)
+        {
             LinkedList sockElem = LinkedList_get(self->sockets, i);
 
-            if (sockElem) {
-                Socket sock = (Socket) LinkedList_getData(sockElem);
+            if (sockElem)
+            {
+                Socket sock = (Socket)LinkedList_getData(sockElem);
 
-                if (sock) {
+                if (sock)
+                {
                     self->fds[i].fd = sock->fd;
                     self->fds[i].events = POLL_IN;
                 }
@@ -134,21 +148,25 @@ Handleset_waitReady(HandleSet self, unsigned int timeoutMs)
         self->pollfdIsUpdated = true;
     }
 
-    if (self->fds && self->nfds > 0) {
+    if (self->fds && self->nfds > 0)
+    {
         int result = poll(self->fds, self->nfds, timeoutMs);
 
-        if (result == -1 && errno == EINTR) {
+        if (result == -1 && errno == EINTR)
+        {
             result = 0;
         }
 
-        if (result == -1) {
+        if (result == -1)
+        {
             if (DEBUG_SOCKET)
                 printf("SOCKET: poll error (errno: %i)\n", errno);
         }
 
         return result;
     }
-    else {
+    else
+    {
         /* there is no socket to wait for */
         return 0;
     }
@@ -157,7 +175,8 @@ Handleset_waitReady(HandleSet self, unsigned int timeoutMs)
 void
 Handleset_destroy(HandleSet self)
 {
-    if (self) {
+    if (self)
+    {
         if (self->sockets)
             LinkedList_destroyStatic(self->sockets);
 
@@ -177,26 +196,30 @@ Socket_activateTcpKeepAlive(Socket self, int idleTime, int interval, int count)
 
     optval = 1;
 
-    if (setsockopt(self->fd, SOL_SOCKET, SO_KEEPALIVE, &optval, optlen)) {
+    if (setsockopt(self->fd, SOL_SOCKET, SO_KEEPALIVE, &optval, optlen))
+    {
         if (DEBUG_SOCKET)
             printf("SOCKET: Failed to enable TCP keepalive\n");
     }
 
 #if defined TCP_KEEPCNT
     optval = idleTime;
-    if (setsockopt(self->fd, IPPROTO_TCP, TCP_KEEPIDLE, &optval, optlen)) {
+    if (setsockopt(self->fd, IPPROTO_TCP, TCP_KEEPIDLE, &optval, optlen))
+    {
         if (DEBUG_SOCKET)
             printf("SOCKET: Failed to set TCP keepalive TCP_KEEPIDLE parameter\n");
     }
 
     optval = interval;
-    if (setsockopt(self->fd, IPPROTO_TCP, TCP_KEEPINTVL, &optval, optlen)) {
+    if (setsockopt(self->fd, IPPROTO_TCP, TCP_KEEPINTVL, &optval, optlen))
+    {
         if (DEBUG_SOCKET)
             printf("SOCKET: Failed to set TCP keepalive TCP_KEEPINTVL parameter\n");
     }
 
     optval = count;
-    if (setsockopt(self->fd, IPPROTO_TCP, TCP_KEEPCNT, &optval, optlen)) {
+    if (setsockopt(self->fd, IPPROTO_TCP, TCP_KEEPCNT, &optval, optlen))
+    {
         if (DEBUG_SOCKET)
             printf("SOCKET: Failed to set TCP keepalive TCP_KEEPCNT parameter\n");
     }
@@ -210,18 +233,20 @@ prepareAddress(const char* address, int port, struct sockaddr_in* sockaddr)
 {
     bool retVal = true;
 
-    memset((char *) sockaddr, 0, sizeof(struct sockaddr_in));
+    memset((char*)sockaddr, 0, sizeof(struct sockaddr_in));
 
-    if (address != NULL) {
+    if (address != NULL)
+    {
         struct addrinfo addressHints;
-        struct addrinfo *lookupResult;
+        struct addrinfo* lookupResult;
         int result;
 
         memset(&addressHints, 0, sizeof(struct addrinfo));
         addressHints.ai_family = AF_INET;
         result = getaddrinfo(address, NULL, &addressHints, &lookupResult);
 
-        if (result != 0) {
+        if (result != 0)
+        {
 
             if (DEBUG_SOCKET)
                 printf("SOCKET: getaddrinfo failed (code=%i)\n", result);
@@ -259,7 +284,7 @@ activateTcpNoDelay(Socket self)
 {
     /* activate TCP_NODELAY option - packets will be sent immediately */
     int flag = 1;
-    setsockopt(self->fd, IPPROTO_TCP, TCP_NODELAY, (char *) &flag, sizeof(int));
+    setsockopt(self->fd, IPPROTO_TCP, TCP_NODELAY, (char*)&flag, sizeof(int));
 }
 
 ServerSocket
@@ -269,22 +294,25 @@ TcpServerSocket_create(const char* address, int port)
 
     int fd;
 
-    if ((fd = socket(AF_INET, SOCK_STREAM, 0)) >= 0) {
+    if ((fd = socket(AF_INET, SOCK_STREAM, 0)) >= 0)
+    {
         struct sockaddr_in serverAddress;
 
-        if (!prepareAddress(address, port, &serverAddress)) {
+        if (!prepareAddress(address, port, &serverAddress))
+        {
             close(fd);
             return NULL;
         }
 
         int optionReuseAddr = 1;
-        setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, (char *) &optionReuseAddr, sizeof(int));
+        setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, (char*)&optionReuseAddr, sizeof(int));
 
 #if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 37)
         int tcpUserTimeout = 10000;
-        int result = setsockopt(fd, SOL_TCP,  TCP_USER_TIMEOUT, &tcpUserTimeout, sizeof(tcpUserTimeout));
+        int result = setsockopt(fd, SOL_TCP, TCP_USER_TIMEOUT, &tcpUserTimeout, sizeof(tcpUserTimeout));
 
-        if (result < 0) {
+        if (result < 0)
+        {
             if (DEBUG_SOCKET)
                 printf("SOCKET: failed to set TCP_USER_TIMEOUT\n");
         }
@@ -292,16 +320,18 @@ TcpServerSocket_create(const char* address, int port)
 #warning "TCP_USER_TIMEOUT not supported by linux kernel"
 #endif
 
-        if (bind(fd, (struct sockaddr *) &serverAddress, sizeof(serverAddress)) >= 0) {
-            serverSocket = (ServerSocket) GLOBAL_MALLOC(sizeof(struct sServerSocket));
+        if (bind(fd, (struct sockaddr*)&serverAddress, sizeof(serverAddress)) >= 0)
+        {
+            serverSocket = (ServerSocket)GLOBAL_MALLOC(sizeof(struct sServerSocket));
             serverSocket->fd = fd;
             serverSocket->backLog = 2;
 
-            setSocketNonBlocking((Socket) serverSocket);
+            setSocketNonBlocking((Socket)serverSocket);
         }
-        else {
+        else
+        {
             close(fd);
-            return NULL ;
+            return NULL;
         }
     }
 
@@ -311,7 +341,8 @@ TcpServerSocket_create(const char* address, int port)
 void
 ServerSocket_listen(ServerSocket self)
 {
-    if (listen(self->fd, self->backLog) == -1) {
+    if (listen(self->fd, self->backLog) == -1)
+    {
         if (DEBUG_SOCKET)
             printf("SOCKET: listen failed (errno: %i)\n", errno);
     }
@@ -325,19 +356,22 @@ ServerSocket_accept(ServerSocket self)
 
     Socket conSocket = NULL;
 
-    fd = accept(self->fd, NULL, NULL );
+    fd = accept(self->fd, NULL, NULL);
 
-    if (fd >= 0) {
-        conSocket = (Socket) GLOBAL_CALLOC(1, sizeof(struct sSocket));
+    if (fd >= 0)
+    {
+        conSocket = (Socket)GLOBAL_CALLOC(1, sizeof(struct sSocket));
 
-        if (conSocket) {
+        if (conSocket)
+        {
             conSocket->fd = fd;
 
             setSocketNonBlocking(conSocket);
 
             activateTcpNoDelay(conSocket);
         }
-        else {
+        else
+        {
             /* out of memory */
             close(fd);
 
@@ -345,7 +379,8 @@ ServerSocket_accept(ServerSocket self)
                 printf("SOCKET: out of memory\n");
         }
     }
-    else {
+    else
+    {
         if (DEBUG_SOCKET)
             printf("SOCKET: accept failed (errno=%i)\n", errno);
     }
@@ -362,7 +397,8 @@ ServerSocket_setBacklog(ServerSocket self, int backlog)
 static void
 closeAndShutdownSocket(int socketFd)
 {
-    if (socketFd != -1) {
+    if (socketFd != -1)
+    {
 
         if (DEBUG_SOCKET)
             printf("SOCKET: call shutdown for %i!\n", socketFd);
@@ -370,14 +406,16 @@ closeAndShutdownSocket(int socketFd)
         /* shutdown is required to unblock read or accept in another thread! */
         int result = shutdown(socketFd, SHUT_RDWR);
 
-        if (result == -1) {
+        if (result == -1)
+        {
             if (DEBUG_SOCKET)
                 printf("SOCKET: shutdown error: %i\n", errno);
         }
 
         result = close(socketFd);
 
-        if (result == -1) {
+        if (result == -1)
+        {
             if (DEBUG_SOCKET)
                 printf("SOCKET: close error: %i\n", errno);
         }
@@ -405,24 +443,28 @@ TcpSocket_create()
 
     int sock = socket(AF_INET, SOCK_STREAM, 0);
 
-    if (sock != -1) {
-        self = (Socket) GLOBAL_MALLOC(sizeof(struct sSocket));
+    if (sock != -1)
+    {
+        self = (Socket)GLOBAL_MALLOC(sizeof(struct sSocket));
 
-        if (self) {
+        if (self)
+        {
             self->fd = sock;
             self->connectTimeout = 5000;
 
 #if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 37)
             int tcpUserTimeout = 10000;
-            int result = setsockopt(sock, SOL_TCP,  TCP_USER_TIMEOUT, &tcpUserTimeout, sizeof(tcpUserTimeout));
+            int result = setsockopt(sock, SOL_TCP, TCP_USER_TIMEOUT, &tcpUserTimeout, sizeof(tcpUserTimeout));
 
-            if (result == -1) {
+            if (result == -1)
+            {
                 if (DEBUG_SOCKET)
                     printf("SOCKET: failed to set TCP_USER_TIMEOUT (errno=%i)\n", errno);
             }
 #endif
         }
-        else {
+        else
+        {
             /* out of memory */
             close(sock);
 
@@ -430,7 +472,8 @@ TcpSocket_create()
                 printf("SOCKET: out of memory\n");
         }
     }
-    else {
+    else
+    {
         if (DEBUG_SOCKET)
             printf("SOCKET: failed to create socket (errno=%i)\n", errno);
     }
@@ -454,7 +497,8 @@ Socket_bind(Socket self, const char* srcAddress, int srcPort)
 
     int result = bind(self->fd, (struct sockaddr*)&localAddress, sizeof(localAddress));
 
-    if (result == -1) {
+    if (result == -1)
+    {
         if (DEBUG_SOCKET)
             printf("SOCKET: failed to bind TCP socket (errno=%i)\n", errno);
 
@@ -462,7 +506,7 @@ Socket_bind(Socket self, const char* srcAddress, int srcPort)
         self->fd = -1;
 
         return false;
-    }    
+    }
 
     return true;
 }
@@ -478,18 +522,17 @@ Socket_connectAsync(Socket self, const char* address, int port)
     if (!prepareAddress(address, port, &serverAddress))
         return false;
 
-    fd_set fdSet;
-    FD_ZERO(&fdSet);
-    FD_SET(self->fd, &fdSet);
-
     activateTcpNoDelay(self);
 
     fcntl(self->fd, F_SETFL, O_NONBLOCK);
 
-    if (connect(self->fd, (struct sockaddr *) &serverAddress, sizeof(serverAddress)) < 0) {
+    if (connect(self->fd, (struct sockaddr*)&serverAddress, sizeof(serverAddress)) < 0)
+    {
 
-        if (errno != EINPROGRESS) {
-            if (close(self->fd) == -1) {
+        if (errno != EINPROGRESS)
+        {
+            if (close(self->fd) == -1)
+            {
                 if (DEBUG_SOCKET)
                     printf("SOCKET: failed to close socket (errno: %i)\n", errno);
             }
@@ -505,24 +548,23 @@ Socket_connectAsync(Socket self, const char* address, int port)
 SocketState
 Socket_checkAsyncConnectState(Socket self)
 {
-    struct timeval timeout;
-    timeout.tv_sec = 0;
-    timeout.tv_usec = 0;
+    struct pollfd fds[1];
 
-    fd_set fdSet;
-    FD_ZERO(&fdSet);
-    FD_SET(self->fd, &fdSet);
+    fds[0].fd = self->fd;
+    fds[0].events = POLLOUT;
 
-    int selectVal = select(self->fd + 1, NULL, &fdSet , NULL, &timeout);
+    int result = poll(fds, 1, 0);
 
-    if (selectVal == 1) {
+    if (result == 1)
+    {
 
         /* Check if connection is established */
 
         int so_error;
         socklen_t len = sizeof so_error;
 
-        if (getsockopt(self->fd, SOL_SOCKET, SO_ERROR, &so_error, &len) >= 0) {
+        if (getsockopt(self->fd, SOL_SOCKET, SO_ERROR, &so_error, &len) >= 0)
+        {
 
             if (so_error == 0)
                 return SOCKET_STATE_CONNECTED;
@@ -530,10 +572,12 @@ Socket_checkAsyncConnectState(Socket self)
 
         return SOCKET_STATE_FAILED;
     }
-    else if (selectVal == 0) {
+    else if (result == 0)
+    {
         return SOCKET_STATE_CONNECTING;
     }
-    else {
+    else
+    {
         return SOCKET_STATE_FAILED;
     }
 }
@@ -544,29 +588,30 @@ Socket_connect(Socket self, const char* address, int port)
     if (Socket_connectAsync(self, address, port) == false)
         return false;
 
-    struct timeval timeout;
-    timeout.tv_sec = self->connectTimeout / 1000;
-    timeout.tv_usec = (self->connectTimeout % 1000) * 1000;
+    struct pollfd fds[1];
 
-    fd_set fdSet;
-    FD_ZERO(&fdSet);
-    FD_SET(self->fd, &fdSet);
+    fds[0].fd = self->fd;
+    fds[0].events = POLLOUT;
 
-    if (select(self->fd + 1, NULL, &fdSet , NULL, &timeout) == 1) {
+    int result = poll(fds, 1, self->connectTimeout);
+
+    if (result == 1)
+    {
 
         /* Check if connection is established */
 
         int so_error;
         socklen_t len = sizeof so_error;
 
-        if (getsockopt(self->fd, SOL_SOCKET, SO_ERROR, &so_error, &len) >= 0) {
+        if (getsockopt(self->fd, SOL_SOCKET, SO_ERROR, &so_error, &len) >= 0)
+        {
 
             if (so_error == 0)
                 return true;
         }
     }
 
-    close (self->fd);
+    close(self->fd);
     self->fd = -1;
 
     return false;
@@ -580,22 +625,24 @@ convertAddressToStr(struct sockaddr_storage* addr)
 
     bool isIPv6;
 
-    if (addr->ss_family == AF_INET) {
-        struct sockaddr_in* ipv4Addr = (struct sockaddr_in*) addr;
+    if (addr->ss_family == AF_INET)
+    {
+        struct sockaddr_in* ipv4Addr = (struct sockaddr_in*)addr;
         port = ntohs(ipv4Addr->sin_port);
         inet_ntop(AF_INET, &(ipv4Addr->sin_addr), addrString, INET_ADDRSTRLEN);
         isIPv6 = false;
     }
-    else if (addr->ss_family == AF_INET6) {
-        struct sockaddr_in6* ipv6Addr = (struct sockaddr_in6*) addr;
+    else if (addr->ss_family == AF_INET6)
+    {
+        struct sockaddr_in6* ipv6Addr = (struct sockaddr_in6*)addr;
         port = ntohs(ipv6Addr->sin6_port);
         inet_ntop(AF_INET6, &(ipv6Addr->sin6_addr), addrString, INET6_ADDRSTRLEN);
         isIPv6 = true;
     }
     else
-        return NULL ;
+        return NULL;
 
-    char* clientConnection = (char*) GLOBAL_MALLOC(strlen(addrString) + 9);
+    char* clientConnection = (char*)GLOBAL_MALLOC(strlen(addrString) + 9);
 
     if (isIPv6)
         sprintf(clientConnection, "[%s]:%i", addrString, port);
@@ -611,7 +658,8 @@ Socket_getPeerAddress(Socket self)
     struct sockaddr_storage addr;
     socklen_t addrLen = sizeof(addr);
 
-    if (getpeername(self->fd, (struct sockaddr*) &addr, &addrLen) == 0) {
+    if (getpeername(self->fd, (struct sockaddr*)&addr, &addrLen) == 0)
+    {
         return convertAddressToStr(&addr);
     }
     else
@@ -624,7 +672,8 @@ Socket_getLocalAddress(Socket self)
     struct sockaddr_storage addr;
     socklen_t addrLen = sizeof(addr);
 
-    if (getsockname(self->fd, (struct sockaddr*) &addr, &addrLen) == 0) {
+    if (getsockname(self->fd, (struct sockaddr*)&addr, &addrLen) == 0)
+    {
         return convertAddressToStr(&addr);
     }
     else
@@ -636,28 +685,37 @@ Socket_getPeerAddressStatic(Socket self, char* peerAddressString)
 {
     struct sockaddr_storage addr;
     socklen_t addrLen = sizeof(addr);
+    memset(&addr, 0, sizeof(addr));
 
-    getpeername(self->fd, (struct sockaddr*) &addr, &addrLen);
+    if (getpeername(self->fd, (struct sockaddr*)&addr, &addrLen) == -1)
+    {
+        if (DEBUG_SOCKET)
+            printf("DEBUG_SOCKET: getpeername -> errno: %i\n", errno);
+
+        return NULL;
+    }
 
     char addrString[INET6_ADDRSTRLEN + 7];
     int port;
 
     bool isIPv6;
 
-    if (addr.ss_family == AF_INET) {
-        struct sockaddr_in* ipv4Addr = (struct sockaddr_in*) &addr;
+    if (addr.ss_family == AF_INET)
+    {
+        struct sockaddr_in* ipv4Addr = (struct sockaddr_in*)&addr;
         port = ntohs(ipv4Addr->sin_port);
         inet_ntop(AF_INET, &(ipv4Addr->sin_addr), addrString, INET_ADDRSTRLEN);
         isIPv6 = false;
     }
-    else if (addr.ss_family == AF_INET6) {
-        struct sockaddr_in6* ipv6Addr = (struct sockaddr_in6*) &addr;
+    else if (addr.ss_family == AF_INET6)
+    {
+        struct sockaddr_in6* ipv6Addr = (struct sockaddr_in6*)&addr;
         port = ntohs(ipv6Addr->sin6_port);
         inet_ntop(AF_INET6, &(ipv6Addr->sin6_addr), addrString, INET6_ADDRSTRLEN);
         isIPv6 = true;
     }
     else
-        return NULL ;
+        return NULL;
 
     if (isIPv6)
         sprintf(peerAddressString, "[%s]:%i", addrString, port);
@@ -678,22 +736,24 @@ Socket_read(Socket self, uint8_t* buf, int size)
     if (read_bytes == 0)
         return -1;
 
-    if (read_bytes == -1) {
+    if (read_bytes == -1)
+    {
         int error = errno;
 
-        switch (error) {
+        switch (error)
+        {
 
-            case EAGAIN:
-                return 0;
-            case EBADF:
-                return -1;
+        case EAGAIN:
+            return 0;
+        case EBADF:
+            return -1;
 
-            default:
+        default:
 
-                if (DEBUG_SOCKET)
-                    printf("DEBUG_SOCKET: recv returned error (errno=%i)\n", error);
+            if (DEBUG_SOCKET)
+                printf("DEBUG_SOCKET: recv returned error (errno=%i)\n", error);
 
-                return -1;
+            return -1;
         }
     }
 
@@ -709,11 +769,14 @@ Socket_write(Socket self, uint8_t* buf, int size)
     /* MSG_NOSIGNAL - prevent send to signal SIGPIPE when peer unexpectedly closed the socket */
     int retVal = send(self->fd, buf, size, MSG_NOSIGNAL | MSG_DONTWAIT);
 
-    if (retVal == -1) {
-        if (errno == EAGAIN) {
+    if (retVal == -1)
+    {
+        if (errno == EAGAIN)
+        {
             return 0;
         }
-        else {
+        else
+        {
             if (DEBUG_SOCKET)
                 printf("DEBUG_SOCKET: send returned error (errno=%i)\n", errno);
         }
@@ -736,19 +799,32 @@ Socket_destroy(Socket self)
     GLOBAL_FREEMEM(self);
 }
 
-UdpSocket
-UdpSocket_create()
+static UdpSocket
+UdpSocket_createUsingNamespace(int namespace)
 {
     UdpSocket self = NULL;
 
-    int sock = socket(AF_INET, SOCK_DGRAM, 0);
+    int sock = socket(namespace, SOCK_DGRAM, IPPROTO_UDP);
 
-    if (sock != -1) {
-        self = (UdpSocket) GLOBAL_MALLOC(sizeof(struct sSocket));
+    if (sock != -1)
+    {
+        self = (UdpSocket)GLOBAL_MALLOC(sizeof(struct sSocket));
 
-        self->fd = sock;
+        if (self)
+        {
+            self->fd = sock;
+            self->namespace = namespace;
+        }
+        else
+        {
+            if (DEBUG_SOCKET)
+                printf("SOCKET: failed to allocate memory\n");
+
+            close(sock);
+        }
     }
-    else {
+    else
+    {
         if (DEBUG_SOCKET)
             printf("SOCKET: failed to create UDP socket (errno=%i)\n", errno);
     }
@@ -756,12 +832,102 @@ UdpSocket_create()
     return self;
 }
 
+UdpSocket
+UdpSocket_create()
+{
+    return UdpSocket_createUsingNamespace(AF_INET);
+}
+
+UdpSocket
+UdpSocket_createIpV6()
+{
+    return UdpSocket_createUsingNamespace(AF_INET6);
+}
+
+bool
+UdpSocket_addGroupMembership(UdpSocket self, const char* multicastAddress)
+{
+    if (self->namespace == AF_INET)
+    {
+        struct ip_mreq mreq;
+
+        if (!inet_aton(multicastAddress, &(mreq.imr_multiaddr)))
+        {
+            printf("SOCKET: Invalid IPv4 multicast address\n");
+            return false;
+        }
+        else
+        {
+            mreq.imr_interface.s_addr = htonl(INADDR_ANY);
+
+            if (setsockopt(self->fd, IPPROTO_IP, IP_ADD_MEMBERSHIP, &mreq, sizeof(mreq)) == -1)
+            {
+                printf("SOCKET: failed to set IPv4 multicast group (errno: %i)\n", errno);
+                return false;
+            }
+        }
+
+        return true;
+    }
+    else if (self->namespace == AF_INET6)
+    {
+        struct ipv6_mreq mreq;
+
+        if (inet_pton(AF_INET6, multicastAddress, &(mreq.ipv6mr_multiaddr)) < 1)
+        {
+            printf("SOCKET: failed to set IPv6 multicast group (errno: %i)\n", errno);
+            return false;
+        }
+
+        mreq.ipv6mr_interface = 0;
+
+        if (setsockopt(self->fd, IPPROTO_IPV6, IPV6_ADD_MEMBERSHIP, &mreq, sizeof(mreq)) == -1)
+        {
+            printf("SOCKET: failed to set IPv6 multicast group (errno: %i)\n", errno);
+            return false;
+        }
+
+        return true;
+    }
+
+    return false;
+}
+
+bool
+UdpSocket_setMulticastTtl(UdpSocket self, int ttl)
+{
+    if (self->namespace == AF_INET)
+    {
+        if (setsockopt(self->fd, IPPROTO_IP, IP_MULTICAST_TTL, &ttl, sizeof(ttl)) == -1)
+        {
+            printf("SOCKET: failed to set IPv4 multicast TTL (errno: %i)\n", errno);
+            return false;
+        }
+
+        return true;
+    }
+    else if (self->namespace == AF_INET6)
+    {
+        if (setsockopt(self->fd, IPPROTO_IPV6, IPV6_MULTICAST_HOPS, &ttl, sizeof(ttl)) == -1)
+        {
+            printf("SOCKET: failed to set IPv6 multicast TTL(hops) (errno: %i)\n", errno);
+            return false;
+        }
+
+        return true;
+    }
+
+    return false;
+}
+
 bool
 UdpSocket_bind(UdpSocket self, const char* address, int port)
 {
+    // TODO add support for IPv6
     struct sockaddr_in localAddress;
 
-    if (!prepareAddress(address, port, &localAddress)) {
+    if (!prepareAddress(address, port, &localAddress))
+    {
         close(self->fd);
         self->fd = 0;
         return false;
@@ -769,7 +935,8 @@ UdpSocket_bind(UdpSocket self, const char* address, int port)
 
     int result = bind(self->fd, (struct sockaddr*)&localAddress, sizeof(localAddress));
 
-    if (result == -1) {
+    if (result == -1)
+    {
         if (DEBUG_SOCKET)
             printf("SOCKET: failed to bind UDP socket (errno=%i)\n", errno);
 
@@ -785,9 +952,11 @@ UdpSocket_bind(UdpSocket self, const char* address, int port)
 bool
 UdpSocket_sendTo(UdpSocket self, const char* address, int port, uint8_t* msg, int msgSize)
 {
+    // TODO add support for IPv6
     struct sockaddr_in remoteAddress;
 
-    if (!prepareAddress(address, port, &remoteAddress)) {
+    if (!prepareAddress(address, port, &remoteAddress))
+    {
 
         if (DEBUG_SOCKET)
             printf("SOCKET: failed to lookup remote address %s\n", address);
@@ -797,14 +966,17 @@ UdpSocket_sendTo(UdpSocket self, const char* address, int port, uint8_t* msg, in
 
     int result = sendto(self->fd, msg, msgSize, 0, (struct sockaddr*)&remoteAddress, sizeof(remoteAddress));
 
-    if (result == msgSize) {
+    if (result == msgSize)
+    {
         return true;
     }
-    else if (result == -1) {
+    else if (result == -1)
+    {
         if (DEBUG_SOCKET)
             printf("SOCKET: failed to send UDP message (errno=%i)\n", errno);
     }
-    else {
+    else
+    {
         if (DEBUG_SOCKET)
             printf("SOCKET: failed to send UDP message (insufficient data sent)\n");
     }
@@ -815,35 +987,41 @@ UdpSocket_sendTo(UdpSocket self, const char* address, int port, uint8_t* msg, in
 int
 UdpSocket_receiveFrom(UdpSocket self, char* address, int maxAddrSize, uint8_t* msg, int msgSize)
 {
+    // TODO add support for IPv6
     struct sockaddr_storage remoteAddress;
+    memset(&remoteAddress, 0, sizeof(struct sockaddr_storage));
     socklen_t structSize = sizeof(struct sockaddr_storage);
 
     int result = recvfrom(self->fd, msg, msgSize, MSG_DONTWAIT, (struct sockaddr*)&remoteAddress, &structSize);
 
-    if (result == -1) {
+    if (result == -1)
+    {
         if (DEBUG_SOCKET)
             printf("SOCKET: failed to receive UDP message (errno=%i)\n", errno);
     }
 
-    if (address) {
+    if (address)
+    {
         bool isIPv6;
         char addrString[INET6_ADDRSTRLEN + 7];
         int port;
 
-        if (remoteAddress.ss_family == AF_INET) {
-            struct sockaddr_in* ipv4Addr = (struct sockaddr_in*) &remoteAddress;
+        if (remoteAddress.ss_family == AF_INET)
+        {
+            struct sockaddr_in* ipv4Addr = (struct sockaddr_in*)&remoteAddress;
             port = ntohs(ipv4Addr->sin_port);
             inet_ntop(AF_INET, &(ipv4Addr->sin_addr), addrString, INET_ADDRSTRLEN);
             isIPv6 = false;
         }
-        else if (remoteAddress.ss_family == AF_INET6) {
-            struct sockaddr_in6* ipv6Addr = (struct sockaddr_in6*) &remoteAddress;
+        else if (remoteAddress.ss_family == AF_INET6)
+        {
+            struct sockaddr_in6* ipv6Addr = (struct sockaddr_in6*)&remoteAddress;
             port = ntohs(ipv6Addr->sin6_port);
             inet_ntop(AF_INET6, &(ipv6Addr->sin6_addr), addrString, INET6_ADDRSTRLEN);
             isIPv6 = true;
         }
         else
-            return result ;
+            return result;
 
         if (isIPv6)
             snprintf(address, maxAddrSize, "[%s]:%i", addrString, port);
