@@ -1,7 +1,7 @@
 /*
  *  time.c
  *
- *  Copyright 2013-2021 Michael Zillgith
+ *  Copyright 2013-2024 Michael Zillgith
  *
  *  This file is part of Platform Abstraction Layer (libpal)
  *  for libiec61850, libmms, and lib60870.
@@ -9,19 +9,6 @@
 
 #include "hal_time.h"
 #include <time.h>
-
-#ifdef CONFIG_SYSTEM_HAS_CLOCK_GETTIME
-uint64_t
-Hal_getTimeInMs()
-{
-	struct timespec tp;
-
-	clock_gettime(CLOCK_REALTIME, &tp);
-
-	return ((uint64_t) tp.tv_sec) * 1000LL + (tp.tv_nsec / 1000000);
-}
-#else
-
 #include <sys/time.h>
 
 msSinceEpoch
@@ -41,8 +28,8 @@ Hal_getTimeInNs()
 
     clock_gettime(CLOCK_REALTIME, &now);
 
-    nsSinceEpoch nsTime = now.tv_sec * 1000000000UL;
-    nsTime += now.tv_nsec;
+    nsSinceEpoch nsTime = (nsSinceEpoch)(now.tv_sec) * 1000000000UL;
+    nsTime += (nsSinceEpoch)(now.tv_nsec);
 
     return nsTime;
 }
@@ -62,6 +49,33 @@ Hal_setTimeInNs(nsSinceEpoch nsTime)
     return true;
 }
 
+msSinceEpoch
+Hal_getMonotonicTimeInMs()
+{
+    uint64_t timeVal = 0;
 
-#endif
+    struct timespec ts;
 
+    if (clock_gettime (CLOCK_MONOTONIC, &ts) == 0)
+    {
+        timeVal = ((uint64_t)ts.tv_sec * 1000LL) + (ts.tv_nsec / 1000000);
+    }
+
+    return timeVal;
+}
+
+nsSinceEpoch
+Hal_getMonotonicTimeInNs()
+{
+    uint64_t nsTime = 0;
+
+    struct timespec ts;
+
+    if (clock_gettime(CLOCK_REALTIME, &ts) == 0)
+    {
+        nsTime = ts.tv_sec * 1000000000UL;
+        nsTime += ts.tv_nsec;
+    }
+
+    return nsTime;
+}
