@@ -175,16 +175,20 @@ CS101_ASDU_encode(CS101_ASDU self, Frame frame)
 }
 
 CS101_ASDU
-CS101_ASDU_createFromBuffer(CS101_AppLayerParameters parameters, uint8_t* msg, int msgLength)
+CS101_ASDU_createFromBufferEx(CS101_ASDU asdu, CS101_AppLayerParameters parameters, uint8_t* msg, int msgLength)
 {
     int asduHeaderLength = 2 + parameters->sizeOfCOT + parameters->sizeOfCA;
 
     if (msgLength < asduHeaderLength)
         return NULL;
 
-    CS101_ASDU self = (CS101_ASDU) GLOBAL_MALLOC(sizeof(struct sCS101_ASDU));
+    CS101_ASDU self = (CS101_ASDU)asdu;
 
-    if (self != NULL) {
+    if (self == NULL)
+        self = (CS101_ASDU) GLOBAL_MALLOC(sizeof(struct sCS101_ASDU));
+
+    if (self)
+    {
         self->parameters = parameters;
 
         self->asdu = msg;
@@ -197,6 +201,12 @@ CS101_ASDU_createFromBuffer(CS101_AppLayerParameters parameters, uint8_t* msg, i
     return self;
 }
 
+CS101_ASDU
+CS101_ASDU_createFromBuffer(CS101_AppLayerParameters parameters, uint8_t* msg, int msgLength)
+{
+    return CS101_ASDU_createFromBufferEx(NULL, parameters, msg, msgLength);
+}
+
 uint8_t*
 CS101_ASDU_getPayload(CS101_ASDU self)
 {
